@@ -1,47 +1,13 @@
 /* pngquant.c - quantize the colors in an alphamap down to a specified number
 **
-** Copyright (C) 1989, 1991 by Jef Poskanzer.
-**
-** Permission to use, copy, modify, and distribute this software and its
-** documentation for any purpose and without fee is hereby granted, provided
-** that the above copyright notice appear in all copies and that both that
-** copyright notice and this permission notice appear in supporting
-** documentation.  This software is provided "as is" without express or
-** implied warranty.
-**
-** - - - -
-**
+** © 2009-2019 by Kornel Lesiński.
+** © 1989, 1991 by Jef Poskanzer.
 ** © 1997-2002 by Greg Roelofs; based on an idea by Stefan Schneider.
-** © 2009-2015 by Kornel Lesiński.
-**
-** All rights reserved.
-**
-** Redistribution and use in source and binary forms, with or without modification,
-** are permitted provided that the following conditions are met:
-**
-** 1. Redistributions of source code must retain the above copyright notice,
-**    this list of conditions and the following disclaimer.
-**
-** 2. Redistributions in binary form must reproduce the above copyright notice,
-**    this list of conditions and the following disclaimer in the documentation
-**    and/or other materials provided with the distribution.
-**
-** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
-** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
-** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
-** SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
-** CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-** OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **
+** See COPYRIGHT file for license.
 */
 
-#define PNGQUANT_VERSION LIQ_VERSION_STRING " (March 2017)"
-
-#define PNGQUANT_USAGE "\
+char *PNGQUANT_USAGE = "\
 usage:  pngquant [options] [ncolors] -- pngfile [pngfile ...]\n\
         pngquant [options] [ncolors] - >stdout <stdin\n\n\
 options:\n\
@@ -50,7 +16,7 @@ options:\n\
   --output file     destination file path to use instead of --ext (synonym: -o)\n\
   --ext new.png     set custom suffix/extension for output filenames\n\
   --quality min-max don't save below min, use fewer colors below max (0-100)\n\
-  --speed N         speed/quality trade-off. 1=slow, 3=default, 11=fast & rough\n\
+  --speed N         speed/quality trade-off. 1=slow, 4=default, 11=fast & rough\n\
   --nofs            disable Floyd-Steinberg dithering\n\
   --posterize N     output lower-precision color (e.g. for ARGB4444 output)\n\
   --strip           remove optional metadata (default on Mac)\n\
@@ -60,8 +26,10 @@ Quantizes one or more 32-bit RGBA PNGs to 8-bit (or smaller) RGBA-palette.\n\
 The output filename is the same as the input name except that\n\
 it ends in \"-fs8.png\", \"-or8.png\" or your custom extension (unless the\n\
 input is stdin, in which case the quantized image will go to stdout).\n\
+If you pass the special output path \"-\" and a single input file, that file\n\
+will be processed and the quantized image will go to stdout.\n\
 The default behavior if the output file exists is to skip the conversion;\n\
-use --force to overwrite. See man page for full list of options.\n"
+use --force to overwrite. See man page for full list of options.\n";
 
 
 #include <stdio.h>
@@ -69,16 +37,13 @@ use --force to overwrite. See man page for full list of options.\n"
 #include <string.h>
 #include <stdarg.h>
 #include <stdbool.h>
-#include <getopt.h>
-#include <unistd.h>
 #include <math.h>
 
-extern char *optarg;
-extern int optind, opterr;
-
-#if defined(WIN32) || defined(__WIN32__)
+#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
 #  include <fcntl.h>    /* O_BINARY */
 #  include <io.h>   /* setmode() */
+#else
+#  include <unistd.h>
 #endif
 
 #ifdef _OPENMP
@@ -89,28 +54,19 @@ extern int optind, opterr;
 #endif
 
 #include "rwpng.h"  /* typedefs, common macros, public prototypes */
-#include "lib/libimagequant.h"
+#include "libimagequant.h" /* if it fails here, run: git submodule update; ./configure; or add -Ilib to compiler flags */
+#include "pngquant_opts.h"
 
-struct pngquant_options {
-    liq_attr *liq;
-    liq_image *fixed_palette_image;
-    liq_log_callback_function *log_callback;
-    void *log_callback_user_info;
-    float floyd;
-    bool using_stdin, using_stdout, force, fast_compression, ie_mode,
-        min_quality_limit, skip_if_larger,
-        strip,
-        verbose;
-};
+char *PNGQUANT_VERSION = LIQ_VERSION_STRING " (July 2019)";
 
 static pngquant_error prepare_output_image(liq_result *result, liq_image *input_image, rwpng_color_transform tag, png8_image *output_image);
 static void set_palette(liq_result *result, png8_image *output_image);
 static pngquant_error read_image(liq_attr *options, const char *filename, int using_stdin, png24_image *input_image_p, liq_image **liq_image_p, bool keep_input_pixels, bool strip, bool verbose);
-static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options);
+static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options, liq_attr *liq);
 static char *add_filename_extension(const char *filename, const char *newext);
 static bool file_exists(const char *outname);
 
-static void verbose_printf(struct pngquant_options *context, const char *fmt, ...)
+static void verbose_printf(liq_attr *liq, struct pngquant_options *context, const char *fmt, ...)
 {
     if (context->log_callback) {
         va_list va;
@@ -118,12 +74,19 @@ static void verbose_printf(struct pngquant_options *context, const char *fmt, ..
         int required_space = vsnprintf(NULL, 0, fmt, va)+1; // +\0
         va_end(va);
 
+#if defined(_MSC_VER)
+        char *buf = malloc(required_space);
+#else
         char buf[required_space];
+#endif
         va_start(va, fmt);
         vsnprintf(buf, required_space, fmt, va);
         va_end(va);
 
-        context->log_callback(context->liq, buf, context->log_callback_user_info);
+        context->log_callback(liq, buf, context->log_callback_user_info);
+#if defined(_MSC_VER)
+        free(buf);
+#endif
     }
 }
 
@@ -163,9 +126,8 @@ static void log_callback_buferred(const liq_attr *attr, const char *msg, void* c
 }
 #endif
 
-static void print_full_version(FILE *fd)
-{
-    fprintf(fd, "pngquant, %s, by Kornel Lesinski, Greg Roelofs.\n"
+void pngquant_internal_print_config(FILE *fd) {
+    fputs(""
         #ifndef NDEBUG
                     "   WARNING: this is a DEBUG (slow) version.\n" /* NDEBUG disables assert() */
         #endif
@@ -175,7 +137,21 @@ static void print_full_version(FILE *fd)
         #if _OPENMP
                     "   Compiled with OpenMP (multicore support).\n"
         #endif
-        , PNGQUANT_VERSION);
+    , fd);
+    fflush(fd);
+}
+
+FILE *pngquant_c_stderr() {
+    return stderr;
+}
+FILE *pngquant_c_stdout() {
+    return stdout;
+}
+
+static void print_full_version(FILE *fd)
+{
+    fprintf(fd, "pngquant, %s, by Kornel Lesinski, Greg Roelofs.\n", PNGQUANT_VERSION);
+    pngquant_internal_print_config(fd);
     rwpng_version_info(fd);
     fputs("\n", fd);
 }
@@ -222,260 +198,148 @@ static bool parse_quality(const char *quality, liq_attr *options, bool *min_qual
     return LIQ_OK == liq_set_quality(options, limit, target);
 }
 
-static const struct {const char *old; const char *newopt;} obsolete_options[] = {
-    {"-fs","--floyd=1"},
-    {"-nofs", "--ordered"},
-    {"-floyd", "--floyd=1"},
-    {"-nofloyd", "--ordered"},
-    {"-ordered", "--ordered"},
-    {"-force", "--force"},
-    {"-noforce", "--no-force"},
-    {"-verbose", "--verbose"},
-    {"-quiet", "--quiet"},
-    {"-noverbose", "--quiet"},
-    {"-noquiet", "--verbose"},
-    {"-help", "--help"},
-    {"-version", "--version"},
-    {"-ext", "--ext"},
-    {"-speed", "--speed"},
-};
-
-static void fix_obsolete_options(const unsigned int argc, char *argv[])
-{
-    for(unsigned int argn=1; argn < argc; argn++) {
-        if ('-' != argv[argn][0]) continue;
-
-        if ('-' == argv[argn][1]) break; // stop on first --option or --
-
-        for(unsigned int i=0; i < sizeof(obsolete_options)/sizeof(obsolete_options[0]); i++) {
-            if (0 == strcmp(obsolete_options[i].old, argv[argn])) {
-                fprintf(stderr, "  warning: option '%s' has been replaced with '%s'.\n", obsolete_options[i].old, obsolete_options[i].newopt);
-                argv[argn] = (char*)obsolete_options[i].newopt;
-            }
-        }
-    }
-}
-
-enum {arg_floyd=1, arg_ordered, arg_ext, arg_no_force, arg_iebug,
-    arg_transbug, arg_map, arg_posterize, arg_skip_larger, arg_strip};
-
-static const struct option long_options[] = {
-    {"verbose", no_argument, NULL, 'v'},
-    {"quiet", no_argument, NULL, 'q'},
-    {"force", no_argument, NULL, 'f'},
-    {"no-force", no_argument, NULL, arg_no_force},
-    {"floyd", optional_argument, NULL, arg_floyd},
-    {"ordered", no_argument, NULL, arg_ordered},
-    {"nofs", no_argument, NULL, arg_ordered},
-    {"iebug", no_argument, NULL, arg_iebug},
-    {"transbug", no_argument, NULL, arg_transbug},
-    {"ext", required_argument, NULL, arg_ext},
-    {"skip-if-larger", no_argument, NULL, arg_skip_larger},
-    {"output", required_argument, NULL, 'o'},
-    {"speed", required_argument, NULL, 's'},
-    {"quality", required_argument, NULL, 'Q'},
-    {"posterize", required_argument, NULL, arg_posterize},
-    {"strip", no_argument, NULL, arg_strip},
-    {"map", required_argument, NULL, arg_map},
-    {"version", no_argument, NULL, 'V'},
-    {"help", no_argument, NULL, 'h'},
-    {NULL, 0, NULL, 0},
-};
-
-pngquant_error pngquant_file(const char *filename, const char *outname, struct pngquant_options *options);
-
+pngquant_error pngquant_main_internal(struct pngquant_options *options, liq_attr *liq);
+static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options, liq_attr *liq);
 
+#ifndef PNGQUANT_NO_MAIN
 int main(int argc, char *argv[])
 {
     struct pngquant_options options = {
         .floyd = 1.f, // floyd-steinberg dithering
         .strip = false,
     };
-    options.liq = liq_attr_create();
 
-    if (!options.liq) {
+    pngquant_error retval = pngquant_parse_options(argc, argv, &options);
+    if (retval != SUCCESS) {
+        return retval;
+    }
+
+    if (options.print_version) {
+        puts(PNGQUANT_VERSION);
+        return SUCCESS;
+    }
+
+    if (options.missing_arguments) {
+        print_full_version(stderr);
+        print_usage(stderr);
+        return MISSING_ARGUMENT;
+    }
+
+    if (options.print_help) {
+        print_full_version(stdout);
+        print_usage(stdout);
+        return SUCCESS;
+    }
+
+    liq_attr *liq = liq_attr_create();
+
+    if (!liq) {
         fputs("SSE-capable CPU is required for this build.\n", stderr);
         return WRONG_ARCHITECTURE;
     }
 
-    unsigned int error_count=0, skipped_count=0, file_count=0;
-    pngquant_error latest_error=SUCCESS;
-    const char *newext = NULL, *output_file_path = NULL;
+    if (options.quality && !parse_quality(options.quality, liq, &options.min_quality_limit)) {
+        fputs("Quality should be in format min-max where min and max are numbers in range 0-100.\n", stderr);
+        return INVALID_ARGUMENT;
+    }
 
-    fix_obsolete_options(argc, argv);
+    if (options.iebug) {
+        // opacities above 238 will be rounded up to 255, because IE6 truncates <255 to 0.
+        liq_set_min_opacity(liq, 238);
+        fputs("  warning: the workaround for IE6 is deprecated\n", stderr);
+    }
 
-    int opt;
-    do {
-        opt = getopt_long(argc, argv, "Vvqfhs:Q:o:", long_options, NULL);
-        switch (opt) {
-            case 'v':
-                options.verbose = true;
-                break;
-            case 'q':
-                options.verbose = false;
-                break;
+    if (options.verbose) {
+        liq_set_log_callback(liq, log_callback, NULL);
+        options.log_callback = log_callback;
+    }
 
-            case arg_floyd:
-                options.floyd = optarg ? atof(optarg) : 1.f;
-                if (options.floyd < 0 || options.floyd > 1.f) {
-                    fputs("--floyd argument must be in 0..1 range\n", stderr);
-                    return INVALID_ARGUMENT;
-                }
-                break;
-            case arg_ordered: options.floyd = 0; break;
+    if (options.last_index_transparent) {
+        liq_set_last_index_transparent(liq, true);
+    }
 
-            case 'f': options.force = true; break;
-            case arg_no_force: options.force = false; break;
-
-            case arg_ext: newext = optarg; break;
-            case 'o':
-                if (output_file_path) {
-                    fputs("--output option can be used only once\n", stderr);
-                    return INVALID_ARGUMENT;
-                }
-                output_file_path = optarg; break;
-
-            case arg_iebug:
-                // opacities above 238 will be rounded up to 255, because IE6 truncates <255 to 0.
-                liq_set_min_opacity(options.liq, 238);
-                fputs("  warning: the workaround for IE6 is deprecated\n", stderr);
-                break;
-
-            case arg_transbug:
-                liq_set_last_index_transparent(options.liq, true);
-                break;
-
-            case arg_skip_larger:
-                options.skip_if_larger = true;
-                break;
-
-            case 's':
-                {
-                    int speed = atoi(optarg);
-                    if (speed >= 10) {
-                        options.fast_compression = true;
-                    }
-                    if (speed == 11) {
-                        options.floyd = 0;
-                        speed = 10;
-                    }
-                    if (LIQ_OK != liq_set_speed(options.liq, speed)) {
-                        fputs("Speed should be between 1 (slow) and 11 (fast).\n", stderr);
-                        return INVALID_ARGUMENT;
-                    }
-                }
-                break;
-
-            case 'Q':
-                if (!parse_quality(optarg, options.liq, &options.min_quality_limit)) {
-                    fputs("Quality should be in format min-max where min and max are numbers in range 0-100.\n", stderr);
-                    return INVALID_ARGUMENT;
-                }
-                break;
-
-            case arg_posterize:
-                if (LIQ_OK != liq_set_min_posterization(options.liq, atoi(optarg))) {
-                    fputs("Posterization should be number of bits in range 0-4.\n", stderr);
-                    return INVALID_ARGUMENT;
-                }
-                break;
-
-            case arg_strip:
-                options.strip = true;
-                break;
-
-            case arg_map:
-                {
-                    png24_image tmp = {};
-                    if (SUCCESS != read_image(options.liq, optarg, false, &tmp, &options.fixed_palette_image, true, true, false)) {
-                        fprintf(stderr, "  error: unable to load %s", optarg);
-                        return INVALID_ARGUMENT;
-                    }
-                    liq_result *tmp_quantize = liq_quantize_image(options.liq, options.fixed_palette_image);
-                    const liq_palette *pal = liq_get_palette(tmp_quantize);
-                    if (!pal) {
-                        fprintf(stderr, "  error: unable to read colors from %s", optarg);
-                        return INVALID_ARGUMENT;
-                    }
-                    for(unsigned int i=0; i < pal->count; i++) {
-                        liq_image_add_fixed_color(options.fixed_palette_image, pal->entries[i]);
-                    }
-                    liq_result_destroy(tmp_quantize);
-                }
-                break;
-
-            case 'h':
-                print_full_version(stdout);
-                print_usage(stdout);
-                return SUCCESS;
-
-            case 'V':
-                puts(PNGQUANT_VERSION);
-                return SUCCESS;
-
-            case -1: break;
-
-            default:
-                return INVALID_ARGUMENT;
+    if (options.speed >= 10) {
+        options.fast_compression = true;
+        if (options.speed == 11) {
+            options.floyd = 0;
+            options.speed = 10;
         }
-    } while (opt != -1);
+    }
 
-    int argn = optind;
+    if (options.speed && LIQ_OK != liq_set_speed(liq, options.speed)) {
+        fputs("Speed should be between 1 (slow) and 11 (fast).\n", stderr);
+        return INVALID_ARGUMENT;
+    }
 
-    if (argn >= argc) {
-        if (argn > 1) {
-            fputs("No input files specified.\n", stderr);
-        } else {
+    if (options.colors && LIQ_OK != liq_set_max_colors(liq, options.colors)) {
+        fputs("Number of colors must be between 2 and 256.\n", stderr);
+        return INVALID_ARGUMENT;
+    }
+
+    if (options.posterize && LIQ_OK != liq_set_min_posterization(liq, options.posterize)) {
+        fputs("Posterization should be number of bits in range 0-4.\n", stderr);
+        return INVALID_ARGUMENT;
+    }
+
+    if (options.extension && options.output_file_path) {
+        fputs("--ext and --output options can't be used at the same time\n", stderr);
+        return INVALID_ARGUMENT;
+    }
+
+    // new filename extension depends on options used. Typically basename-fs8.png
+    if (options.extension == NULL) {
+        options.extension = options.floyd > 0 ? "-fs8.png" : "-or8.png";
+    }
+
+    if (options.output_file_path && options.num_files != 1) {
+        fputs("  error: Only one input file is allowed when --output is used. This error also happens when filenames with spaces are not in quotes.\n", stderr);
+        return INVALID_ARGUMENT;
+    }
+
+    if (options.using_stdout && !options.using_stdin && options.num_files != 1) {
+        fputs("  error: Only one input file is allowed when using the special output path \"-\" to write to stdout. This error also happens when filenames with spaces are not in quotes.\n", stderr);
+        return INVALID_ARGUMENT;
+    }
+
+    if (!options.num_files && !options.using_stdin) {
+        fputs("No input files specified.\n", stderr);
+        if (options.verbose) {
             print_full_version(stderr);
         }
         print_usage(stderr);
         return MISSING_ARGUMENT;
     }
 
-    if (options.verbose) {
-        liq_set_log_callback(options.liq, log_callback, NULL);
-        options.log_callback = log_callback;
-    }
+    retval = pngquant_main_internal(&options, liq);
+    liq_attr_destroy(liq);
+    return retval;
+}
+#endif
 
-    char *colors_end;
-    unsigned long colors = strtoul(argv[argn], &colors_end, 10);
-    if (colors_end != argv[argn] && '\0' == colors_end[0]) {
-        if (LIQ_OK != liq_set_max_colors(options.liq, colors)) {
-            fputs("Number of colors must be between 2 and 256.\n", stderr);
+// Don't use this. This is not a public API.
+pngquant_error pngquant_main_internal(struct pngquant_options *options, liq_attr *liq)
+{
+    if (options->map_file) {
+        png24_image tmp = {.width=0};
+        if (SUCCESS != read_image(liq, options->map_file, false, &tmp, &options->fixed_palette_image, true, true, false)) {
+            fprintf(stderr, "  error: unable to load %s", options->map_file);
             return INVALID_ARGUMENT;
         }
-        argn++;
-    }
-
-    if (newext && output_file_path) {
-        fputs("--ext and --output options can't be used at the same time\n", stderr);
-        return INVALID_ARGUMENT;
-    }
-
-    // new filename extension depends on options used. Typically basename-fs8.png
-    if (newext == NULL) {
-        newext = options.floyd > 0 ? "-ie-fs8.png" : "-ie-or8.png";
-        if (!options.ie_mode) {
-            newext += 3;    /* skip "-ie" */
+        liq_result *tmp_quantize = liq_quantize_image(liq, options->fixed_palette_image);
+        const liq_palette *pal = liq_get_palette(tmp_quantize);
+        if (!pal) {
+            fprintf(stderr, "  error: unable to read colors from %s", options->map_file);
+            return INVALID_ARGUMENT;
         }
-    }
-
-    if (argn == argc || (argn == argc-1 && 0==strcmp(argv[argn],"-"))) {
-        options.using_stdin = true;
-        options.using_stdout = !output_file_path;
-        argn = argc-1;
-    }
-
-    const int num_files = argc-argn;
-
-    if (output_file_path && num_files != 1) {
-        fputs("Only one input file is allowed when --output is used\n", stderr);
-        return INVALID_ARGUMENT;
+        for(unsigned int i=0; i < pal->count; i++) {
+            liq_image_add_fixed_color(options->fixed_palette_image, pal->entries[i]);
+        }
+        liq_result_destroy(tmp_quantize);
     }
 
 #ifdef _OPENMP
     // if there's a lot of files, coarse parallelism can be used
-    if (num_files > 2*omp_get_max_threads()) {
+    if (options->num_files > 2*omp_get_max_threads()) {
         omp_set_nested(0);
         omp_set_dynamic(1);
     } else {
@@ -483,19 +347,22 @@ int main(int argc, char *argv[])
     }
 #endif
 
+    unsigned int error_count=0, skipped_count=0, file_count=0;
+    pngquant_error latest_error=SUCCESS;
+
     #pragma omp parallel for \
         schedule(static, 1) reduction(+:skipped_count) reduction(+:error_count) reduction(+:file_count) shared(latest_error)
-    for(int i=0; i < num_files; i++) {
-        struct pngquant_options opts = options;
-        opts.liq = liq_attr_copy(options.liq);
+    for(int i=0; i < options->num_files; i++) {
+        const char *filename = options->using_stdin ? "stdin" : options->files[i];
+        struct pngquant_options opts = *options;
+        liq_attr *local_liq = liq_attr_copy(liq);
 
-        const char *filename = opts.using_stdin ? "stdin" : argv[argn+i];
 
         #ifdef _OPENMP
-        struct buffered_log buf = {};
-        if (opts.log_callback && omp_get_num_threads() > 1 && num_files > 1) {
-            liq_set_log_callback(opts.liq, log_callback_buferred, &buf);
-            liq_set_log_flush_callback(opts.liq, log_callback_buferred_flush, &buf);
+        struct buffered_log buf = {0};
+        if (opts.log_callback && omp_get_num_threads() > 1 && opts.num_files > 1) {
+            liq_set_log_callback(local_liq, log_callback_buferred, &buf);
+            liq_set_log_flush_callback(local_liq, log_callback_buferred_flush, &buf);
             opts.log_callback = log_callback_buferred;
             opts.log_callback_user_info = &buf;
         }
@@ -504,11 +371,11 @@ int main(int argc, char *argv[])
 
         pngquant_error retval = SUCCESS;
 
-        const char *outname = output_file_path;
+        const char *outname = opts.output_file_path;
         char *outname_free = NULL;
         if (!opts.using_stdout) {
             if (!outname) {
-                outname = outname_free = add_filename_extension(filename, newext);
+                outname = outname_free = add_filename_extension(filename, opts.extension);
             }
             if (!opts.force && file_exists(outname)) {
                 fprintf(stderr, "  error: '%s' exists; not overwriting\n", outname);
@@ -517,12 +384,12 @@ int main(int argc, char *argv[])
         }
 
         if (SUCCESS == retval) {
-            retval = pngquant_file(filename, outname, &opts);
+            retval = pngquant_file_internal(filename, outname, &opts, local_liq);
         }
 
         free(outname_free);
 
-        liq_attr_destroy(opts.liq);
+        liq_attr_destroy(local_liq);
 
         if (retval) {
             #pragma omp critical
@@ -539,60 +406,60 @@ int main(int argc, char *argv[])
     }
 
     if (error_count) {
-        verbose_printf(&options, "There were errors quantizing %d file%s out of a total of %d file%s.",
+        verbose_printf(liq, options, "There were errors quantizing %d file%s out of a total of %d file%s.",
                        error_count, (error_count == 1)? "" : "s", file_count, (file_count == 1)? "" : "s");
     }
     if (skipped_count) {
-        verbose_printf(&options, "Skipped %d file%s out of a total of %d file%s.",
+        verbose_printf(liq, options, "Skipped %d file%s out of a total of %d file%s.",
                        skipped_count, (skipped_count == 1)? "" : "s", file_count, (file_count == 1)? "" : "s");
     }
     if (!skipped_count && !error_count) {
-        verbose_printf(&options, "Quantized %d image%s.",
+        verbose_printf(liq, options, "Quantized %d image%s.",
                        file_count, (file_count == 1)? "" : "s");
     }
 
-    if (options.fixed_palette_image) liq_image_destroy(options.fixed_palette_image);
-    liq_attr_destroy(options.liq);
+    if (options->fixed_palette_image) liq_image_destroy(options->fixed_palette_image);
 
     return latest_error;
 }
 
-pngquant_error pngquant_file(const char *filename, const char *outname, struct pngquant_options *options)
+/// Don't hack this. Instead use https://github.com/ImageOptim/libimagequant/blob/f54d2f1a3e1cf728e17326f4db0d45811c63f063/example.c
+static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options, liq_attr *liq)
 {
     pngquant_error retval = SUCCESS;
 
-    verbose_printf(options, "%s:", filename);
+    verbose_printf(liq, options, "%s:", filename);
 
     liq_image *input_image = NULL;
-    png24_image input_image_rwpng = {};
+    png24_image input_image_rwpng = {.width=0};
     bool keep_input_pixels = options->skip_if_larger || (options->using_stdout && options->min_quality_limit); // original may need to be output to stdout
     if (SUCCESS == retval) {
-        retval = read_image(options->liq, filename, options->using_stdin, &input_image_rwpng, &input_image, keep_input_pixels, options->strip, options->verbose);
+        retval = read_image(liq, filename, options->using_stdin, &input_image_rwpng, &input_image, keep_input_pixels, options->strip, options->verbose);
     }
 
     int quality_percent = 90; // quality on 0-100 scale, updated upon successful remap
-    png8_image output_image = {};
+    png8_image output_image = {.width=0};
     if (SUCCESS == retval) {
-        verbose_printf(options, "  read %luKB file", (input_image_rwpng.file_size+1023UL)/1024UL);
+        verbose_printf(liq, options, "  read %luKB file", (input_image_rwpng.file_size+1023UL)/1024UL);
 
         if (RWPNG_ICCP == input_image_rwpng.input_color) {
-            verbose_printf(options, "  used embedded ICC profile to transform image to sRGB colorspace");
+            verbose_printf(liq, options, "  used embedded ICC profile to transform image to sRGB colorspace");
         } else if (RWPNG_GAMA_CHRM == input_image_rwpng.input_color) {
-            verbose_printf(options, "  used gAMA and cHRM chunks to transform image to sRGB colorspace");
+            verbose_printf(liq, options, "  used gAMA and cHRM chunks to transform image to sRGB colorspace");
         } else if (RWPNG_ICCP_WARN_GRAY == input_image_rwpng.input_color) {
-            verbose_printf(options, "  warning: ignored ICC profile in GRAY colorspace");
+            verbose_printf(liq, options, "  warning: ignored ICC profile in GRAY colorspace");
         } else if (RWPNG_COCOA == input_image_rwpng.input_color) {
             // No comment
         } else if (RWPNG_SRGB == input_image_rwpng.input_color) {
-            verbose_printf(options, "  passing sRGB tag from the input");
+            verbose_printf(liq, options, "  passing sRGB tag from the input");
         } else if (input_image_rwpng.gamma != 0.45455) {
-            verbose_printf(options, "  converted image from gamma %2.1f to gamma 2.2",
+            verbose_printf(liq, options, "  converted image from gamma %2.1f to gamma 2.2",
                            1.0/input_image_rwpng.gamma);
         }
 
         // when using image as source of a fixed palette the palette is extracted using regular quantization
         liq_result *remap;
-        liq_error remap_error = liq_image_quantize(options->fixed_palette_image ? options->fixed_palette_image : input_image, options->liq, &remap);
+        liq_error remap_error = liq_image_quantize(options->fixed_palette_image ? options->fixed_palette_image : input_image, liq, &remap);
 
         if (LIQ_OK == remap_error) {
 
@@ -612,7 +479,7 @@ pngquant_error pngquant_file(const char *filename, const char *outname, struct p
                 double palette_error = liq_get_quantization_error(remap);
                 if (palette_error >= 0) {
                     quality_percent = liq_get_quantization_quality(remap);
-                    verbose_printf(options, "  mapped image to new colors...MSE=%.3f (Q=%d)", palette_error, quality_percent);
+                    verbose_printf(liq, options, "  mapped image to new colors...MSE=%.3f (Q=%d)", palette_error, quality_percent);
                 }
             }
             liq_result_destroy(remap);
@@ -628,7 +495,7 @@ pngquant_error pngquant_file(const char *filename, const char *outname, struct p
         if (options->skip_if_larger) {
             // this is very rough approximation, but generally avoid losing more quality than is gained in file size.
             // Quality is raised to 1.5, because even greater savings are needed to justify big quality loss.
-            // but >50% savings are considered always worthwile in order to allow low quality conversions to work at all
+            // but >50% savings are considered always worthwhile in order to allow low quality conversions to work at all
             const double quality = quality_percent/100.0;
             const double expected_reduced_size = pow(quality, 1.5);
             output_image.maximum_file_size = (input_image_rwpng.file_size-1) * (expected_reduced_size < 0.5 ? 0.5 : expected_reduced_size);
@@ -636,20 +503,20 @@ pngquant_error pngquant_file(const char *filename, const char *outname, struct p
 
         output_image.fast_compression = options->fast_compression;
         output_image.chunks = input_image_rwpng.chunks; input_image_rwpng.chunks = NULL;
-        retval = write_image(&output_image, NULL, outname, options);
+        retval = write_image(&output_image, NULL, outname, options, liq);
 
         if (TOO_LARGE_FILE == retval) {
-            verbose_printf(options, "  file exceeded expected size of %luKB", (unsigned long)output_image.maximum_file_size/1024UL);
+            verbose_printf(liq, options, "  file exceeded expected size of %luKB", (unsigned long)output_image.maximum_file_size/1024UL);
         }
         if (SUCCESS == retval && output_image.metadata_size > 0) {
-            verbose_printf(options, "  copied %dKB of additional PNG metadata", (int)(output_image.metadata_size+999)/1000);
+            verbose_printf(liq, options, "  copied %dKB of additional PNG metadata", (int)(output_image.metadata_size+999)/1000);
         }
     }
 
     if (options->using_stdout && keep_input_pixels && (TOO_LARGE_FILE == retval || TOO_LOW_QUALITY == retval)) {
         // when outputting to stdout it'd be nasty to create 0-byte file
         // so if quality is too low, output 24-bit original
-        pngquant_error write_retval = write_image(NULL, &input_image_rwpng, outname, options);
+        pngquant_error write_retval = write_image(NULL, &input_image_rwpng, outname, options, liq);
         if (write_retval) {
             retval = write_retval;
         }
@@ -694,8 +561,8 @@ static char *add_filename_extension(const char *filename, const char *newext)
     char* outname = malloc(x+4+strlen(newext)+1);
     if (!outname) return NULL;
 
-    strncpy(outname, filename, x);
-    if (strncmp(outname+x-4, ".png", 4) == 0 || strncmp(outname+x-4, ".PNG", 4) == 0) {
+    strcpy(outname, filename);
+    if (x > 4 && (strncmp(outname+x-4, ".png", 4) == 0 || strncmp(outname+x-4, ".PNG", 4) == 0)) {
         strcpy(outname+x-4, newext);
     } else {
         strcpy(outname+x, newext);
@@ -718,7 +585,7 @@ static char *temp_filename(const char *basename) {
 
 static void set_binary_mode(FILE *fp)
 {
-#if defined(WIN32) || defined(__WIN32__)
+#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
     setmode(fp == stdout ? 1 : 0, O_BINARY);
 #endif
 }
@@ -734,7 +601,7 @@ static const char *filename_part(const char *path)
 }
 
 static bool replace_file(const char *from, const char *to, const bool force) {
-#if defined(WIN32) || defined(__WIN32__)
+#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
     if (force) {
         // On Windows rename doesn't replace
         unlink(to);
@@ -743,7 +610,7 @@ static bool replace_file(const char *from, const char *to, const bool force) {
     return (0 == rename(from, to));
 }
 
-static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options)
+static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options, liq_attr *liq)
 {
     FILE *outfile;
     char *tempname = NULL;
@@ -753,9 +620,9 @@ static pngquant_error write_image(png8_image *output_image, png24_image *output_
         outfile = stdout;
 
         if (output_image) {
-            verbose_printf(options, "  writing %d-color image to stdout", output_image->num_palette);
+            verbose_printf(liq, options, "  writing %d-color image to stdout", output_image->num_palette);
         } else {
-            verbose_printf(options, "  writing truecolor image to stdout");
+            verbose_printf(liq, options, "  writing truecolor image to stdout");
         }
     } else {
         tempname = temp_filename(outname);
@@ -768,9 +635,9 @@ static pngquant_error write_image(png8_image *output_image, png24_image *output_
         }
 
         if (output_image) {
-            verbose_printf(options, "  writing %d-color image as %s", output_image->num_palette, filename_part(outname));
+            verbose_printf(liq, options, "  writing %d-color image as %s", output_image->num_palette, filename_part(outname));
         } else {
-            verbose_printf(options, "  writing truecolor image as %s", filename_part(outname));
+            verbose_printf(liq, options, "  writing truecolor image as %s", filename_part(outname));
         }
     }
 
@@ -863,8 +730,8 @@ static pngquant_error prepare_output_image(liq_result *result, liq_image *input_
     ** Step 3.7 [GRR]: allocate memory for the entire indexed image
     */
 
-    output_image->indexed_data = malloc(output_image->height * output_image->width);
-    output_image->row_pointers = malloc(output_image->height * sizeof(output_image->row_pointers[0]));
+    output_image->indexed_data = malloc((size_t)output_image->height * (size_t)output_image->width);
+    output_image->row_pointers = malloc((size_t)output_image->height * sizeof(output_image->row_pointers[0]));
 
     if (!output_image->indexed_data || !output_image->row_pointers) {
         return OUT_OF_MEMORY_ERROR;
